host = 'davte.it'
host = 'localhost'
port = 5000

certificate = '/home/davte/Projects/filebridging/data/filesharing.crt'
key = '/home/davte/Projects/filebridging/data/filesharing.key'

password = 'pepperoni'
token = 'aaa111ddd'
